from chatmof.llm.llama import get_llama_llm
from chatmof.llm.openai import get_openai_llm
from chatmof.llm.codellama import get_codellama_llm
from langchain.base_language import BaseLanguageModel


def get_llm(model_name: str, temperature: float = 0.1, max_tokens: int = 4096) -> BaseLanguageModel:
    m_name = model_name.lower()

    if m_name.startswith('gpt'):
        llm = get_openai_llm(model_name, temperature=temperature)
    elif m_name.startswith('llama'):
        llm = get_llama_llm(model_name, max_token=max_tokens, temperature=temperature)
    elif m_name.startswith('codellama'):
        llm = get_codellama_llm(model_name, max_token=max_tokens, temperature=temperature)
    else:
        raise ValueError(f'model_name should be one of llama and gpt, not {model_name}')
    
    return llm



import os
os.environ['CURL_CA_BUNDLE'] = ''
os.environ["CUDA_VISIBLE_DEVICES"]='1'

import argparse
import copy
from itertools import chain
import warnings

from langchain.callbacks import StdOutCallbackHandler
from langchain.callbacks.manager import CallbackManagerForChainRun
from chatmof.config import config as config_default
from chatmof.agents.agent import ChatMOF
from chatmof.llm import get_llm

warnings.filterwarnings(action='ignore')


str_kwargs_names = {
    'accelerator': "Device name for MOFTransformer. accelerator must be one of [cuda, gpu, cpu] (default: cuda)",
    'logger': 'logger for generation task (Default: generate_mof.log)',
}

int_kwargs_names = {
    'max_length_in_predictor': 'max number of MOFs in predictor step. (default: 30)',
    'num_genetic_cycle': 'number of genetic algorithm cycle (default: 3)',
}

float_kwargs_names = {
}

bool_kwargs_names = {
    'search_internet': 'If True, using "search internet" tools. (default = False)',
    'verbose': 'If True, print intermediate step. (default = True)'
}


def main(model='gpt-4', temperature=0.1, **kwargs) -> str:
    config = copy.deepcopy(config_default)

    config.update(kwargs)
    search_internet = config['search_internet']
    verbose = config['verbose']

    llm = get_llm(model, temperature=temperature)
    callback_manager = [StdOutCallbackHandler()]

    run_manager = CallbackManagerForChainRun.get_noop_manager()

    chatmof = ChatMOF.from_llm(
        llm=llm, 
        verbose=verbose, 
        search_internet=search_internet,
    )

    print ('#' * 50 + "\n")
    print ('Welcom to ChatMOF!')
    print ("\n" + "#"*10 + ' Question ' + "#"*30)
    print ('Please enter the question below >>')
    question = input()
    
    output = chatmof.run(question, callbacks=callback_manager)

    print ('\n')
    print ("#"*10 + ' Output ' + "#" * 30)
    print (output)
    print ('\n')
    print ('Thanks for using CHATMOF!')

    return output


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    add = parser.add_argument

    add(
        '--model',
        '--model-name',
        '-m',
        type=str,
        help="OpenAI model name. model_name must be one of [gpt-4, gpt-3.5-turbo, gpt-3.5-turbo-16k]. (default: gpt-4)",
        default='gpt-4'
    )
    
    add(
        '--temperature',
        '-t',
        help='Temperature of LLM models. The lower the temperature, the more accurate the answer, and the higher the temperature, the more variable the answer. (default: 0.1)',
        type=float,
        default=0.1,
    )

    for key, value in str_kwargs_names.items():
        parser.add_argument(f"--{key}", type=str, required=False, help=f"(optional) {value}")

    for key, value in int_kwargs_names.items():
        parser.add_argument(f"--{key}", type=int, required=False, help=f"(optional) {value}")

    for key, value in float_kwargs_names.items():
        parser.add_argument(f"--{key}", type=float, required=False, help=f"(optional) {value}")

    for key, value in bool_kwargs_names.items():
        parser.add_argument(f"--{key}", action='store_true', required=False, help=f"(optional) {value}")

    args = parser.parse_args()
    model = args.model
    temperature = args.temperature

    kwargs = {}
    for key in chain(str_kwargs_names.keys(), 
                        int_kwargs_names.keys(),
                        float_kwargs_names.keys(),
                        bool_kwargs_names.keys(),
                        ):
        if value := getattr(args, key):
            kwargs[key] = value

    main(model=model, temperature=temperature, **kwargs)



import os
from typing import Any
from langchain.tools import Tool
from langchain.base_language import BaseLanguageModel
from langchain.tools.base import BaseTool

from chatmof.config import config
from chatmof.tools.unit_converter.base import UnitConverter


def _get_unit_converter(
        llm: BaseLanguageModel,
        verbose: bool = False,
        **kwargs: Any) -> BaseTool:

    return Tool(
        name="unit_converter",
        description=(
            "A tool for converting unit. "
            "Ask a question about unit conversion and it will tell you how to convert and what information you need. "
            ""
        ),
        func=UnitConverter.from_llm(
            llm=llm, verbose=verbose
        ).run
    )



import os
from typing import Any
from langchain.tools import Tool
from langchain.base_language import BaseLanguageModel
from langchain.tools.base import BaseTool

from chatmof.config import config
from chatmof.tools.ase_repl.base import ASETool


def _get_ase_repl(
        llm: BaseLanguageModel,
        file_path: str = config['lookup_dir'],
        verbose: bool = False,
        **kwargs: Any) -> BaseTool:

    return Tool(
        name="ase_repl",
        description=(
            "A tools for analyzing materials using the ASE library."
            "It allows you to perform tasks using ASE, such as atom coordinates and cell information"
            "The input must be a detailed full sentence to answer the question."
        ),
        func=ASETool.from_llm(
            llm=llm, verbose=verbose,
        ).run
    )



PROMPT = """You should have a plan for converting units for metal-organic frameworks. The question specifies the unit to be converted from and the desired target unit for the property. 
You need to explain the conversion process and identify any additional information required. To answer the question, you have to fill in the following format:

Question: the input question you must answer
Thought: you should always think about what to do
Unit: unit to be converted from and the desired target unit
Equation: you should explain the conversion process
Information: you should answer the information you need in this process

You must be careful with units. The units must match the final units when they are computed.

Begin!

Question: Change the m^2/g unit for surface area to m^2/cm^3
Thought: Thought: To convert the surface area from m^2/g to m^2/cm^3, we need to convert the mass unit from grams to cubic centimeters. This requires the knowledge of the density of the metal-organic framework.
Unit: The unit to be converted from is m^2/g and the desired target unit is m^2/cm^3.
Equation: Surface area (m^2/cm^3) = Surface area (m^2/g) * Density (g/cm^3)
Information: The additional information required for this conversion is the density of the metal-organic framework in g/cm^3.

Question: Change the mol/kg unit for H2 uptake to cm^3/cm^3
Thought: To convert the H2 uptake from mol/kg to cm^3/cm^3, we need to convert the amount of substance from moles to volume and the mass from kilograms to cubic centimeters. This requires the knowledge of the molar volume of H2 at STP and the density of the metal-organic framework.
Unit: The unit to be converted from is mol/kg and the desired target unit is cm^3/cm^3.
Equation: H2 uptake (cm^3/cm^3) = H2 uptake (mol/kg) * Molar volume of H2 (cm^3/mol) * Density (kg/cm^3)
Information: The additional information required for this conversion is the molar volume of H2 at STP (which is 22.4 L/mol or 22400 cm^3/mol at standard temperature and pressure) and the density of the metal-organic framework in kg/cm^3.

Question: {question}"""



ASE_PROMPT = """You are working with a ase library in Python. The name of the object `ase.atoms` is `atoms`.
You should make a valid python command as input. You must use print the output using the `print` function at the end.

Use the following format:

Question: the input question you must answer
Material: name of material
Convert: convert material to atoms object.
Thought: you should always think about what to do
Input: the valid python code only using the ase library
Observation: the result of python code
... (this Thought/Input/Observation can repeat N times)
Final Thought: you should think about how to answer the question based on your observation
Final Answer: the final answer to the original input question. If you can't answer the question, say `nothing`


Begin!

Question: calculate the cell volume of ABAYOU
Material: ABAYOU
Convert: Convert material to atoms object. I now have to write a code using `atoms` object
Thought: To calculate the cell volume of `atoms`, I need to access the cell information of the atoms object and then use the appropriate method to calculate its volume.
Input: 
``` 
import ase
volume = atoms.get_volume()
print(volume)
```
Observation: 18921.03834202299
Final Thought: The volume of the cell associated with the ase.atoms object atoms was successfully calculated using the get_volume() method from the ASE library.
Final Answer: The cell volume of the ABAYOU is 18921.03834202299 cubic Ångströms.

Question: {question}
{agent_scratchpad}
"""


import os
import re
from pathlib import Path
import pandas as pd
import ase
import ase.io
import ase.visualize
from typing import Dict, Any, List, Optional

from langchain.base_language import BaseLanguageModel
from langchain.chains.base import Chain
from langchain.chains.llm import LLMChain
from langchain.prompts import PromptTemplate
from langchain.callbacks.manager import CallbackManagerForChainRun
from langchain.tools.python.tool import PythonAstREPLTool

from chatmof.config import config
from chatmof.utils import search_file
from chatmof.tools.ase_repl.prompt import ASE_PROMPT


class ASETool(Chain):
    """Tools that search csv using ASE agent"""
    llm_chain: LLMChain
    data_dir: Path = Path(config['structure_dir'])
    input_key: str = 'question'
    output_key: str = 'answer'

    @property
    def input_keys(self) -> List[str]:
        return [self.input_key]
    
    @property
    def output_keys(self) -> List[str]:
        return [self.output_key]
    
    def _parse_output(self, text:str) -> Dict[str, Any]:
        thought = re.search(
            r"(?<!Final )Thought:\s*(.+?)\s*(Observation|Final|Input|Thought|Question|Material|Convert|$)", text, re.DOTALL)
        material = re.search(
            r"Material:\s*(.+?)\s*(Observation|Final|Input|Thought|Question|Material|Convert|$)", text, re.DOTALL)
        convert = re.search(
            r"Convert:\s*(.+?)\s*(Observation|Final|Input|Thought|Question|Material|Convert|$)", text, re.DOTALL)
        input_ = re.search(
            r"Input:\s*(?:```|`)?(.+?)(?:```|`)?\s*(Observation|Final|Input|Thought|Question|Material|Convert|$)", text, re.DOTALL)
        final_thought = re.search(
            r"Final Thought:\s*(.+?)\s*(Observation|Final|Input|Thought|Question|Material|Convert|$)", text, re.DOTALL)
        final_answer = re.search(
            r"Final Answer:\s*(.+?)\s*(Observation|Final|Input|Thought|Question|Material|Convert|$)", text, re.DOTALL)
        
        if (not input_) and (not final_answer):
            raise ValueError(f'unknown format from LLM: {text}')
        
        return {
            'Thought': (thought.group(1) if thought else None),
            'Material': (material.group(1) if material else None),
            'Convert': (convert.group(1) if convert else None),
            'Input': (input_.group(1).strip() if input_ else None),
            'Final Thought' : (final_thought.group(1) if final_thought else None),
            'Final Answer': (final_answer.group(1) if final_answer else None),
        }
    
    def _clear_name(self, text:str) -> str:
        remove_list = ['_clean_h', '_clean', '_charged', '_manual', '_ion_b', '_auto', '_SL', ]
        str_remove_list = r"|".join(remove_list)
        return re.sub(rf"({str_remove_list})", "", text)
    
    # @staticmethod
    def _get_atoms(self, material: str):
        ls_st: str[Path] = search_file(f'{material}*.cif', self.data_dir)
        if not ls_st:
            raise ValueError(f'{material} does not exists.')

        path: Path = ls_st[0]
        atoms = ase.io.read(path)
        return atoms
    
    def _write_log(self, action, text, run_manager):
        run_manager.on_text(f"\n[ASE repl] {action}: ", verbose=self.verbose)
        run_manager.on_text(text, verbose=self.verbose, color='yellow')
    
    def _call(
            self,
            inputs: Dict[str, Any],
            run_manager: Optional[CallbackManagerForChainRun] = None
    ):
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        callbacks = _run_manager.get_child()
        return_observation = inputs.get('return_observation', False)
        
        agent_scratchpad = ''
        max_iteration = config['max_iteration']

        input_ = self._clear_name(inputs[self.input_key])

        for i in range(max_iteration + 1):
            llm_output = self.llm_chain.run(
                question=input_,
                agent_scratchpad = agent_scratchpad,
                callbacks=callbacks,
                stop=['Observation:', 'Question:',]
            )

            if not llm_output.strip():
                agent_scratchpad += 'Thought: '
                llm_output = self.llm_chain.run(
                    question=input_,
                    agent_scratchpad = agent_scratchpad,
                    callbacks=callbacks,
                    stop=['Observation:', 'Question:',]
                )
            
            #if llm_output.endswith('Final Answer: success'):
            if re.search(r'Final Answer: (success|.* above|.* DataFrames?).?$', llm_output):
                thought = f'Final Thought: we have to answer the question `{input_}` using observation\n'
                agent_scratchpad += thought
                llm_output = self.llm_chain.run(
                    question=input_,
                    agent_scratchpad = agent_scratchpad,
                    callbacks=callbacks,
                    stop=['Observation:', 'Question:',]
                )
                llm_output = thought + llm_output

            output = self._parse_output(llm_output)

            if output['Final Answer']:
                if output['Thought']:
                    raise ValueError(llm_output)
                
                self._write_log('Final Thought', output['Final Thought'], run_manager)

                final_answer: str = output['Final Answer']
                check_sentence = ''
                if re.search(r'nothing', final_answer):
                    # please use tool `predictor` to get answer.'
                    final_answer = 'There are no data in ASE repl.'
                elif final_answer.endswith('.'):    
                    final_answer += check_sentence
                else:
                    final_answer = f'The answer for question "{input_}" is {final_answer}.{check_sentence}'
                return {self.output_key: final_answer}
            
            elif i >= max_iteration:
                final_answer = 'There are no data in ASE repl'
                self._write_log('Final Thought', final_answer, run_manager)
                return {self.output_key: final_answer}

            else:
                self._write_log('Material', output['Material'], run_manager)
                self._write_log('Convert', output['Convert'], run_manager)
                self._write_log('Thought', output['Thought'], run_manager)
                self._write_log('Input', output['Input'], run_manager)
                
            if output['Material']:
                atoms = self._get_atoms(output['Material'])

            pytool = PythonAstREPLTool(locals={'atoms': atoms})
            observation = str(pytool.run(output['Input'])).strip()
            
            if return_observation:
                if "\n" in observation:
                    self._write_log('Observation', "\n"+observation, run_manager)
                else:
                    self._write_log('Observation', observation, run_manager)
                return {self.output_key: observation}

            if "\n" in observation:
                self._write_log('Observation', "\n"+observation, run_manager)
            else:
                self._write_log('Observation', observation, run_manager)

            if output['Material'] and output['Convert']:
                agent_scratchpad += 'Material: {}\nConvert: {}\n'\
                    .format(output['Material'], output['Convert'])

            agent_scratchpad += 'Thought: {}\nInput: {} \nObservation: {}\n'\
                .format(output['Thought'], output['Input'], observation)

        raise AssertionError('Code Error! please report to author!')

    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        prompt: str = ASE_PROMPT,
        **kwargs
    ) -> Chain:
        template = PromptTemplate(
            template=prompt,
            input_variables=['question', 'agent_scratchpad'],
        )
        llm_chain = LLMChain(llm=llm, prompt=template)
        return cls(llm_chain=llm_chain, **kwargs)



import torch
from transformers import AutoTokenizer, AutoModelForCausalLM, pipeline
from langchain import HuggingFacePipeline
from langchain.base_language import BaseLanguageModel


LLAMA_HF = [
    'Llama-2-7b'
    'Llama-2-7b-hf'
    'Llama-2-7b-chat',
    'Llama-2-7b-chat-hf',
    'Llama-2-13b',
    'Llama-2-13b-hf',
    'Llama-2-13b-chat',
    'Llama-2-13b-chat-hf',
    'Llama-2-70b',
    'Llama-2-70b-hf',
    'Llama-2-70b-chat',
    'Llama-2-70b-chat-hf',
]


def get_llama_llm(model_name: str, max_token: int =4096, temperature: float=0.1) -> BaseLanguageModel:
    model_name = model_name.capitalize()
    if model_name not in LLAMA_HF:
        raise ValueError(f'model_name should be one of {LLAMA_HF}, not {model_name}')
    model_name = f'meta-llama/{model_name}'

    tokenizer = AutoTokenizer.from_pretrained(model_name,
                                            use_auth_token=True,
                                            )
    model = AutoModelForCausalLM.from_pretrained(model_name,
                                                device_map='auto',
                                                torch_dtype=torch.float16,
                                                use_auth_token=True,
                                                )

    pipe = pipeline("text-generation",
                model=model,
                tokenizer=tokenizer,
                torch_dtype=torch.bfloat16,
                device_map="auto",
                do_sample=True,
                top_k=30,
                max_new_tokens=max_token,
                eos_token_id=tokenizer.eos_token_id
                )

    llm = HuggingFacePipeline(pipeline = pipe, model_kwargs = {'temperature':temperature})
    return llm






class CLICommand:
    """
    Install package <GIRDAY> which calculated energy-grid.

    """

    @staticmethod
    def add_arguments(parser):
        pass

    @staticmethod
    def run(args):
        from moftransformer.utils import install_griday

        install_griday()


import torch
from transformers import LlamaForCausalLM, CodeLlamaTokenizer, pipeline
from langchain import HuggingFacePipeline
from langchain.base_language import BaseLanguageModel


CodeLLAMA_HF = [
    'CodeLlama-7b-Python-hf',
    'CodeLlama-13b-Python-hf',
    'CodeLlama-34b-Python-hf',
]


def get_codellama_llm(model_name: str, max_token: int =4096, temperature: float=0.1) -> BaseLanguageModel:
    if model_name not in CodeLLAMA_HF:
        raise ValueError(f'model_name should be one of {CodeLLAMA_HF}, not {model_name}')
    model_name = f'codellama/{model_name}'

    tokenizer = CodeLlamaTokenizer.from_pretrained(model_name,
                                            use_auth_token=True,
                                            )
    model = LlamaForCausalLM.from_pretrained(model_name,
                                                device_map='auto',
                                                torch_dtype=torch.float16,
                                                use_auth_token=True,
                                                )

    pipe = pipeline("text-generation",
                model=model,
                tokenizer=tokenizer,
                torch_dtype=torch.bfloat16,
                device_map="auto",
                do_sample=True,
                top_k=30,
                max_new_tokens=max_token,
                eos_token_id=tokenizer.eos_token_id,
                pad_token_id=tokenizer.eos_token_id,
                )

    llm = HuggingFacePipeline(pipeline = pipe, model_kwargs = {'temperature':temperature})
    return llm






from langchain.chat_models import ChatOpenAI
from langchain.llms import OpenAI
from langchain.base_language import BaseLanguageModel


CHAT_MODEL = [
    'gpt-4',
    'gpt-4-0613',
    'gpt-4-32k',
    'gpt-4-32k-0613',
    'gpt-4-0314',
    'gpt-4-32k-0314',
    'gpt-3.5-turbo',
    'gpt-3.5-turbo-16k',
    'gpt-3.5-turbo-instruct',
    'gpt-3.5-turbo-0613',
    'gpt-3.5-turbo-16k-0613',
    'gpt-3.5-turbo-0301',
]


MODEL = [
    'text-davinci-003',
    'text-davinci-002',
]


def get_openai_llm(model_name: str, temperature: float=0.1) -> BaseLanguageModel:
    model_name = model_name.lower()
    if model_name in CHAT_MODEL:
        llm = ChatOpenAI(model_name=model_name, temperature=temperature)
    elif model_name in MODEL:
        llm = OpenAI(model_name=model_name, temperature=temperature)

    return llm
